#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Functions to create nanowire networks from text file.
# Testing only, not for production.
# 
# Author: Marcus Kasdorf
# Date:   July 15, 2021

from typing import Dict
import numpy as np
from shapely.geometry import LineString
import networkx as nx

from .line_functions import find_intersects
from .units import get_units
from .nanowire_network import NanowireNetwork

def create_NWN_from_txt(
    filename: str,
    conductance: float = 1.0,
    diameter: float = 1.0,
    resistivity: float = 1.0,
    units: Dict[str, float] = None
) -> NanowireNetwork:
    """
    Create a nanowire network represented by a NetworkX graph. Wires are 
    represented by the graph's vertices, while the wire junctions are 
    represented by the graph's edges.

    The text file input is assumed to be in two columns: x and y.
    Each wire is a pair of rows, one row containing the start point and the
    following containing the end point. The first two pairs of rows are the
    electrodes.

    Parameters
    ----------
    filename : str
        Text file containing the start and end locations of the wires.

    conductance : float, optional
        The junction conductance of the nanowires where they intersect.
        Given in units of (Ron)^-1.

    diameter : float, optional
        The diameter of each nanowire. Given in units of D0.

    resistivity : float, optional
        The resistivity of each nanowire. Given in units of rho0.

    units : dict, optional
        Dictionary of custom base units. Defaults to None which will use the 
        default units given in `units.py`

    Returns
    -------
    NWN : Graph
        The created random nanowire network.

    """
    # Get coordinates from text file
    x, y = np.loadtxt(filename, unpack=True)

    # Convert to LineStrings
    line_list = []
    for i in range(0, len(x), 2):
        line_list.append(LineString([(x[i], y[i]), (x[i+1], y[i+1])]))

    # Find dimensions
    length = np.max(x) - np.min(x)
    width = np.max(y) - np.min(y)
    size = length * width

    # Find density
    wire_num = len(line_list)
    density = wire_num / size

    # Get characteristic units
    units = get_units(units)

    # Create NWN graph
    NWN = NanowireNetwork(
        wire_length = None,
        length = length,
        width = width, 
        size = size,
        wire_density = density, 
        wire_num = wire_num,
        junction_conductance = conductance,
        junction_capacitance = None,
        wire_diameter = diameter,
        wire_resistivity = resistivity,
        electrode_list = [],
        lines = [],
        type = "JDA",
        units = units,
    )

    # Add the wires as nodes to the graph
    for i in range(NWN.graph["wire_num"]):
        NWN.graph["lines"].append(line_list[i])
        if i == 0 or i == 1:
            NWN.add_node((i,), electrode=True)
            NWN.graph["electrode_list"].append((i,))
        else:
            NWN.add_node((i,), electrode=False)
        
    # Find intersects and create the edges (junctions)
    intersect_dict = find_intersects(NWN.graph["lines"])
    NWN.add_edges_from(
        [((key[0],), (key[1],)) for key in intersect_dict.keys()], 
        conductance = conductance,
        capacitance = None,
        type = "junction"
    )
    NWN.graph["loc"] = intersect_dict
    
    # Find junction density
    NWN.graph["junction_density"] = len(intersect_dict) / size

    # Create index lookup
    NWN.graph["node_indices"] = {
        node: ind for ind, node in enumerate(sorted(NWN.nodes()))
    }

    return NWN



import pytest
from pathlib import Path

from randomnwn import *
from randomnwn.fromtext import *
import networkx as nx
from numpy import hypot
from collections import Counter


@pytest.fixture
def NWN_benchmark_JDA():
    # Get benchmark file
    current_path = Path(__file__).parent.resolve()
    benchmark = current_path.joinpath("test_networks/benchmark.txt")

    # Create nanowire network
    units = {"Ron": 20.0, "rho0": 22.63676, "D0": 60.0, "l0": 1.0}
    NWN = create_NWN_from_txt(str(benchmark), units=units)
    return NWN


@pytest.fixture
def NWN_benchmark_MNR(NWN_benchmark_JDA):
    NWN = NWN_benchmark_JDA
    convert_NWN_to_MNR(NWN)
    return NWN


@pytest.fixture
def NWN_test1():
    NWN = create_NWN(size=(8, 5), seed=123)
    add_electrodes(
        NWN, ["left", 2, 1, [-0.5, 0.5]], ["right", 2, 1, [-0.5, 0.5]]
    )
    return NWN


def test_shortest_path():
    NWN = create_NWN(seed=123)
    assert NWN.graph["type"] == "JDA"

    path_len, path = nx.single_source_dijkstra(NWN, (33,), (138,))
    ans = [(33,), (330,), (373,), (622,), (420,), (76,), (21,), (723,), (19,), 
        (232,), (123,), (422,), (308,), (166,), (406,), (53,), (736,), (138,)]

    assert path_len == len(ans) - 1
    assert path == ans


def test_benchmark_network_JDA(NWN_benchmark_JDA):
    # Get benchmark network
    NWN = NWN_benchmark_JDA
    assert NWN.graph["type"] == "JDA"

    # Calculate JDA resistance
    V = 1.0
    sol = solve_network(NWN, (0,), (1,), V)
    R = V / sol[-1]
    R *= NWN.graph["units"]["Ron"]

    # Check for the correct JDA resistance
    R_JDA = 20 + 20 + 1 / (1 / (20 + 20) + 1 / 20)  # 160/3
    assert abs(R - R_JDA) < 1e-8


def test_benchmark_network_MNR(NWN_benchmark_MNR):
    # Get benchmark network
    NWN = NWN_benchmark_MNR
    assert NWN.graph["type"] == "MNR"
    units = NWN.graph["units"]

    # Calculate MNR resistance
    V = 1.0
    sol = solve_network(NWN, (0,), (1,), V)
    R = V / sol[-1]
    R *= units["Ron"]

    # Check for the correct MNR resistance
    const = units["rho0"] / (np.pi/4 * units["D0"]**2) * 1e3
    Rin1 = const * 1.2
    Rin2 = const * hypot(0.3, 0.3)
    Rin3 = const * hypot(1.5, 1.5)
    Rin4 = const * 1.5

    R_MNR = 20 + Rin1 + 20 + Rin2 + \
        1 / (1 / (Rin3 + 20) + 1 / (20 + Rin4 + 20))    # ~74.618
    assert abs(R - R_MNR) < 1e-8


@pytest.mark.parametrize("NWN", ["NWN_benchmark_JDA", "NWN_test1"])
def test_MNR_node_count(NWN, request):
    NWN = request.getfixturevalue(NWN)
    assert NWN.graph["type"] == "JDA"
    convert_NWN_to_MNR(NWN)

    # Number of wire junction edges
    n_wire_junctions = Counter(nx.get_edge_attributes(NWN, "type").values())["junction"]

    # Number of edges connected to an electrode
    n_edges_electrode = len(NWN.edges(NWN.graph["electrode_list"]))

    # Number of connected electrodes
    n_connected_electrodes = len([node for node, deg in NWN.degree(NWN.graph["electrode_list"]) if deg > 0])

    # Number of isolated wires
    n_isolated_wires = len([x for x in nx.connected_components(NWN) if len(x) == 1])

    # Compare the number of nodes obtained via edge and node count
    node_count = 2 * n_wire_junctions \
        - n_edges_electrode \
        + n_connected_electrodes \
        + n_isolated_wires
    
    assert node_count == NWN.number_of_nodes()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Functions to create nanowire networks.
# 
# Author: Marcus Kasdorf
# Date:   July 8, 2024

from __future__ import annotations

import networkx as nx
import numpy as np
from collections import Counter
from scipy.integrate import solve_ivp
from functools import lru_cache

import numpy.typing as npt
from scipy.integrate._ivp.ivp import OdeResult
from typing import Callable, Literal, Any, Optional
from shapely.geometry import LineString, Point
from numbers import Number
from .typing import *

from .units import get_units
from .line_functions import create_line, find_intersects
from .nanowires import convert_NWN_to_MNR
from ._models import (
    resist_func,
    _HP_model_no_decay,
    _HP_model_decay,
    _HP_model_chen,
)


class ParameterNotSetError(Exception):
    """
    Raised when an parameter that needs to used has not been set yet.

    Parameters
    ----------
    param : str
        Name of the parameter that needs to be set.

    message : str, optional
        Explanation of the error.

    """
    def __init__(self, message: str, param: Any | None = None):
        super().__init__(message)
        self.param = param


class NanowireNetwork(nx.Graph):
    """
    Internal nanowire network object. Should not be instantiated directly.
    Use `create_NWN()` instead.

    Parameters
    ----------
    incoming_graph_data : input graph, optional
        Data to initialize graph. Same as networkx.Graph object.

    **attr
        Keyword arguments. Same as networkx.Graph object.

    """
    def __init__(self, incoming_graph_data=None, **attr):
        super().__init__(incoming_graph_data, **attr)
        self._resist_func = None
        self._state_vars: list[str] = []
        self._state_vars_is_set: list[bool] = []
        self._wire_junctions = None

    @property
    def type(self) -> Literal["JDA", "MNR"]:
        return self.graph["type"]

    @property
    def electrodes(self) -> list[JDANode]:
        return self.graph["electrode_list"]

    @property
    def n_electrodes(self) -> int:
        return len(self.electrodes)

    @property
    def n_wires(self) -> int:
        """Number of wires in the network. Does not include electrodes."""
        return self.graph["wire_num"] - self.n_electrodes

    @property
    def n_inner_junctions(self) -> int | None:
        if self.type != "MNR":
            return None

        _ = Counter(nx.get_edge_attributes(self, "type").values())
        return _["inner"]

    @property
    def n_wire_junctions(self) -> int:
        _ = Counter(nx.get_edge_attributes(self, "type").values())
        return _["junction"]

    @property
    def units(self) -> dict[str, float]:
        return self.graph["units"]

    @property
    def lines(self) -> list[LineString]:
        """List of LineStrings representing the nanowires. Includes electrodes."""
        return self.graph["lines"]

    @property
    def wire_density(self) -> float:
        """Wire density in units of (l0)^-2. Does not include electrodes."""
        return self.graph["wire_density"]

    @property
    def loc(self) -> dict[NWNNode, Point]:
        """Dictionary of wire junction locations."""
        return self.graph["loc"]

    def get_index(self, node: NWNNode | list[NWNNode]) -> NWNNodeIndex:
        """Return the unique index of a node in the network."""
        return self.graph["node_indices"][node]

    def get_node(self, index: NWNNodeIndex) -> NWNNode:
        """Return the node corresponding to the index."""
        try:
            return next(k for k, v in self.graph["node_indices"].items() if v == index)
        except StopIteration as e:
            raise ValueError("given index does not have a node") from e

    def get_index_from_edge(self, edge: NWNEdge | list[NWNEdge]) -> NWNEdgeIndex | list[NWNEdgeIndex]:
        """Return the indices of the nodes in the edge as a tuple."""
        if isinstance(edge, list):
            return [tuple(map(self.get_index, e)) for e in edge]
        else:
            return tuple(map(self.get_index, edge))

    def to_MNR(self) -> None:
        convert_NWN_to_MNR(self)

    @property
    def wire_junctions(self) -> list[NWNEdge]:
        """
        Return a list of edges with the "type" attribute set to "junction". 
        Once called, the list is cached. If wires are added, clear the cache
        by deleting the property.

        """
        if self._wire_junctions is None:
            self._wire_junctions = [(u, v) for u, v, d in self.edges(data=True) if d["type"] == "junction"]

        return self._wire_junctions
    
    @wire_junctions.deleter
    def wire_junctions(self) -> None:
        self._wire_junctions = None
        self.wire_junction_indices.cache_clear()

    @lru_cache
    def wire_junction_indices(self) -> tuple[list[NWNNodeIndex], list[NWNNodeIndex]]:
        """
        Return the start and end indices of the wire junctions in the network
        as a tuple of lists.
        
        """
        return np.asarray(self.get_index_from_edge(self.wire_junctions)).T
    
    @property
    def state_vars(self) -> list[str]:
        return self._state_vars
    
    @state_vars.setter
    def state_vars(self, names: list[str]) -> None:
        self._state_vars = names
        self._state_vars_is_set = {name: False for name in names}

    @property
    def resistance_function(
        self
    ) -> Callable[[NanowireNetwork, npt.ArrayLike], npt.ArrayLike]:
        """
        Resistance function of the nanowire network. Should have the calling
        signature `func(NWN, param1, [param2, ...])` where `NWN` is the
        nanowire network, and `param1`, `param2`, etc. are the state variables
        of the memristor element(s).

        You can also pass the string "linear" to choose a default linear
        resistance function based on the state variable w in [0, 1].

        The resistance should be nondimensionalized, i.e. the resistance should
        be in units of Ron.

        """
        return self._resist_func

    @resistance_function.setter
    def resistance_function(
        self, 
        func: str | Callable[[NanowireNetwork, npt.ArrayLike], npt.ArrayLike]
    ) -> None:
        if func == "linear":
            self._resist_func = resist_func
        else:
            self._resist_func = func

    def set_state_var(
        self,
        var_name: str,
        value: npt.ArrayLike,
        edge_list: list[NWNEdge] | None = None,
    ) -> None:
        """
        Set the state variable of the memristors (nanowire network wire 
        junctions) in the network.

        Parameters
        ----------
        var_name : str
            Name of the state variable(s) to set.

        value : ndarray or scalar
            Value to set the state variable(s) to.

        edge_list : list of edges, optional
            List of edges to set the state variable for. If None, all wire
            junction edges will be used. Should be the same length as the value
            array.

        """
        value = np.atleast_1d(value)

        if var_name not in self.state_vars:
            cls = self.__class__
            raise ParameterNotSetError(
                f"'{var_name}' is not in {cls.__qualname__}.state_vars (currently is {self.state_vars})."
                f"\nDid you set it using {cls.__qualname__}.state_vars = ['{var_name}', ...]?"
            )
        
        edge_list = self.wire_junctions if edge_list is None else edge_list

        # Set the state variable for the given edges to the same value...
        if value.size == 1:
            nx.set_edge_attributes(self, {
                edge: {var_name: value[0]} for edge in edge_list
            })

        # or to different values
        elif value.size == len(edge_list):
            nx.set_edge_attributes(self, {
                edge: {var_name: value[i]} for i, edge in enumerate(edge_list)
            })

        else:
            raise ValueError(
                f"Length of value array ({value.size}) does not match the number of edges ({len(edge_list)})."
            )
        
        self._state_vars_is_set[var_name] = True
        
    def get_state_var(self, var_name: str, edge_list: list[NWNEdge] | None = None) -> npt.ArrayLike:
        """
        Get the state variable of the memristors (nanowire network wire 
        junctions) in the network.

        Parameters
        ----------
        var_name : str
            Name of the state variable(s) to get.

        edge_list : list of edges, optional
            List of edges to get the state variable from. If None, all wire
            junction edges will be used.

        Returns
        -------
        ndarray or scalar
            Value of the state variable(s).

        """
        if var_name not in self.state_vars:
            cls = self.__class__
            raise AttributeError(
                f"'{var_name}' is not in {cls.__qualname__}.state_vars (currently is {self.state_vars})."
                f"\nDid you set it using {cls.__qualname__}.state_vars = ['{var_name}', ...]?"
            )

        if edge_list is None:
            edge_list = self.wire_junctions

        try:
            return np.array([self[edge[0]][edge[1]][var_name] for edge in edge_list])
        except KeyError as e:
            raise ParameterNotSetError(f"'{var_name}' has not been set yet using `set_state_var`.") from e

    def update_resistance(
        self, 
        state_var_vals: npt.ArrayLike | list[npt.ArrayLike], 
        edge_list: list[NWNEdge] | None = None
    ) -> None:
        """
        Update the resistance of the nanowire network based on the provided
        state variable values. The resistance function should be set before 
        calling this method.

        Parameters
        ----------
        state_var_vals : ndarray or list of ndarrays
            An array of values of the state variables to use in the resistance 
            function. The should be in the same order as the state variables. 
            If the resistance function takes multiple state variables, pass a 
            list of arrays in the same order as `NanowireNetwork.state_vars`.

        edge_list : list of edges, optional
            List of edges to update the resistance for. If None, all wire
            junction edges will be used. In this case, the length of the
            `state_var_vals` array should be the same as the number of wire
            junctions.

        Returns
        -------
        ndarray
            Array of updated resistance values for each edge in the edge_list.

        """
        if self.resistance_function is None:
            raise ParameterNotSetError("Resistance function attribute must be set before updating resistance.")

        if not isinstance(state_var_vals, list):
            state_var_vals = [state_var_vals]

        if edge_list is None:
            edge_list = self.wire_junctions

        R = self.resistance_function(self, *state_var_vals)
        attrs = {
            edge: {"conductance": 1 / R[i]} for i, edge in enumerate(edge_list)   
        }
        nx.set_edge_attributes(self, attrs)

        return R

    def evolve(
        self,
        model: str | Callable,
        t_eval: npt.NDArray,
        source_node: NWNNode | list[NWNNode] = None,
        drain_node: NWNNode | list[NWNNode] = None,
        voltage_func: Callable[[npt.ArrayLike], npt.ArrayLike] = None,
        window_func: Callable[[npt.ArrayLike], npt.ArrayLike] = None,
        *,
        state_vars: Optional[list[str]] = None,
        args: tuple[Any, ...] = (),
        solver: str = "spsolve",
        spsolve_kwargs: dict = {},
        ivp_options: dict = {},
    ) -> OdeResult:
        """
        Evolve the nanowire network using the given model and state variables.
        Returns the solution of the initial value problem solver from scipy
        at the given time points.

        Parameters
        ----------
        model : str or callable
            Model to use for the evolution. One of ["default", "decay", "chen"]
            can be chosen. If a function is given, it should have the calling
            signature `func(t, y, *args)` where `t` is the time, `y` is the
            state variable(s), and `args` are any additional arguments.

        t_eval : ndarray
            Time points to evaluate the solution at.

        source_node : node or list of nodes, optional
            Source node(s) of the network. Needed if you chose one of the 
            built-in models listed in the `model` parameter.

        drain_node : node or list of nodes, optional
            Drain node(s) of the network. Needed if you chose one of the 
            built-in models listed in the `model` parameter.

        voltage_func : callable, optional
            Function that returns the voltage at the given time. Should have
            the calling signature `func(t)` where `t` is the time. Needed if 
            you chose one of the built-in models listed in the `model` 
            parameter.

        window_func : callable, optional
            Window function for the built-in models. Should have the calling
            signature `func(w)` where `w` is the state variable. Needed if you
            chose one of the built-in models listed in the `model` parameter.

        state_vars : list of str, optional
            List of state variables to evolve. If not provided, all
            state variables in `self.state_vars` will be used.

        args : tuple, optional
            Additional arguments to pass to the model function. Most likely, 
            you will need to provide args for the source node(s), drain node(s) 
            and voltage function.

        solver : str, optional
            Sparse matrix solver function name. Defaults to "spsolve".

        spsolve_kwargs : dict, optional
            Additional keyword arguments to pass to the sparse matrix solver.

        ivp_options : dict, optional
            Additional keyword arguments to pass to the IVP solver.
        
        """
        # Get state variable derivative function
        impl_models = {
            "default": _HP_model_no_decay, 
            "decay": _HP_model_decay, 
            "chen": _HP_model_chen
        }
        if model in impl_models.keys():
            deriv = impl_models[model]
        elif callable(model):
            deriv = model
        else:
            raise NotImplementedError(f"Model '{model}' not found.")
        
        if state_vars is None:
            state_vars = self.state_vars

        # Check if state variables are set
        if not all([self._state_vars_is_set[var] for var in state_vars]):
            raise ParameterNotSetError("Not all state variables have not been set yet.")
        
        # Get initial state variable, if there are more than one, they 
        # will be concatenated.
        y0 = np.hstack([self.get_state_var(var) for var in state_vars])

        # Set the tolerance value for the IVP solver
        if "atol" not in ivp_options.keys():
            ivp_options["atol"] = 1e-7
        if "rtol" not in ivp_options.keys():
            ivp_options["rtol"] = 1e-7

        t_span = (t_eval[0], t_eval[-1])

        # Setup IVP args for the solver
        if model in impl_models.keys():
            edge_list = self.wire_junctions
            start_nodes, end_nodes = np.asarray(
                self.get_index_from_edge(self.wire_junctions)
            ).T
            if not callable(window_func):
                raise ValueError("To use a built-in model, a window function must be provided.")
            args = (
                self, source_node, drain_node, voltage_func, edge_list, 
                start_nodes, end_nodes, window_func, solver, spsolve_kwargs
            )

        # Solve how the state variables change over time
        sol = solve_ivp(
            deriv, t_span, y0, "DOP853", t_eval, args=args, **ivp_options
        )

        # Update the state variables
        split = np.split(sol.y[:, -1], len(state_vars))
        for var, new_vals in zip(state_vars, split):
            self.set_state_var(var, new_vals)

        return sol

    def __repr__(self) -> str:
        d = {
            "Type": self.type,
            "Wires": self.n_wires,
            "Electrodes": self.n_electrodes,
            "Inner-wire junctions": self.n_inner_junctions,
            "Wire junctions": self.n_wire_junctions,
            "Length": f"{self.graph['length'] * self.units['l0']:#.4g} um ({self.graph['length']:#.4g} l0)",
            "Width": f"{self.graph['width'] * self.units['l0']:#.4g} um ({self.graph['width']:#.4g} l0)",
            "Wire Density": f"{self.graph['wire_density'] / self.units['l0']**2:#.4g} um^-2 ({self.graph['wire_density']:#.4g} l0^-2)"
        }
        # Get max key length
        m = max(map(len, list(d.keys())))

        # Create string representation
        s = "\n".join([f"{k:>{m}}: {v}" for k, v in d.items()])
        return s


def create_NWN(
    wire_length: float = (7.0 / 7),
    size: tuple | float = (50.0 / 7),
    density: float = (0.3 * 7**2),
    seed: int = None,
    conductance: float = (0.1 / 0.1),
    capacitance: float = 1000,
    diameter: float = (50.0 / 50.0),
    resistivity: float = (22.6 / 22.6),
    units: dict[str, float] = None
) -> NanowireNetwork:
    """
    Create a nanowire network represented by a NetworkX graph. Wires are 
    represented by the graph's vertices, while the wire junctions are 
    represented by the graph's edges.

    The nanowire network starts in the junction-dominated assumption (JDA), but
    can be converted to the multi-nodal representation (MNR) after creation. 

    The density may not be attainable with the given size, as there can only 
    be a integer number of wires. Thus, the closest density to an integer 
    number of wires is used.

    See `units.py` for the units used by the parameters. 

    Parameters
    ----------
    wire_length : float, optional
        Length of each nanowire. Given in units of l0.

    size : 2-tuple or float
        The size of the nanowire network given in units of l0. If a tuple is
        given, it is assumed to be (x-length, y-length). If a number is passed,
        both dimensions will have the same length. The x direction is labeled
        `length`, while the y direction is labeled `width`.

    density : float, optional
        Density of nanowires in the area determined by the width.
        Given in units of (l0)^-2.

    seed : int, optional
        Seed for random nanowire generation.

    conductance : float, optional
        The junction conductance of the nanowires where they intersect.
        Given in units of (Ron)^-1.

    capacitance : float, optional
        The junction capacitance of the nanowires where they intersect.
        Given in microfarads. (Currently unused)

    diameter : float, optional
        The diameter of each nanowire. Given in units of D0.

    resistivity : float, optional
        The resistivity of each nanowire. Given in units of rho0.

    units : dict, optional
        Dictionary of custom base units. Defaults to None which will use the 
        default units given in `units.py`

    Returns
    -------
    NWN : Graph
        The created random nanowire network.

    """
    # Convert size to length and width, size will be the area
    if isinstance(size, tuple):
        length, width = size
        size = size[0] * size[1]
    elif isinstance(size, Number):
        length, width = size, size
        size = size * size
    else:
        raise ValueError("Invalid size type.")

    # Get closest density with an integer number of wires.
    wire_num = round(size * density)
    density = wire_num / size

    # Get characteristic units
    units = get_units(units)

    # Create NWN graph
    NWN = NanowireNetwork(
        wire_length = wire_length,
        length = length,
        width = width,
        size = size,
        wire_density = density,
        wire_num = wire_num,
        junction_conductance = conductance,
        junction_capacitance = capacitance,
        wire_diameter = diameter,
        wire_resistivity = resistivity,
        electrode_list = [],
        lines = [],
        type = "JDA",
        units = units,
        tau = 0.0,
        epsilon = 0.0,
    )

    # Create seeded random generator for testing
    rng = np.random.default_rng(seed)

    # Add the wires as nodes to the graph
    for i in range(NWN.graph["wire_num"]):
        NWN.graph["lines"].append(create_line(
            NWN.graph["wire_length"],
            xmax = NWN.graph["length"],
            ymax = NWN.graph["width"],
            rng = rng
        ))
        NWN.add_node((i,), electrode=False)

    # Find intersects and create the edges (junctions)
    intersect_dict = find_intersects(NWN.graph["lines"])
    NWN.add_edges_from(
        [((key[0],), (key[1],)) for key in intersect_dict.keys()],
        conductance = conductance,
        capacitance = capacitance,
        type = "junction"
    )
    NWN.graph["loc"] = intersect_dict

    # Find junction density
    NWN.graph["junction_density"] = len(intersect_dict) / size

    # Create index lookup
    NWN.graph["node_indices"] = {
        node: ind for ind, node in enumerate(sorted(NWN.nodes()))
    }

    return NWN


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Functions to evolve nanowire networks over time.
# 
# Author: Marcus Kasdorf
# Date:   July 28, 2021

import numpy as np
import networkx as nx
from scipy.integrate import solve_ivp

from numbers import Number
from typing import Callable, Iterable
from scipy.integrate._ivp.ivp import OdeResult
import numpy.typing as npt

from .nanowire_network import NanowireNetwork
from .typing import *

from .nanowires import get_edge_indices
from .calculations import solve_drain_current, solve_network
from ._models import (
    resist_func,
    _HP_model_no_decay,
    _HP_model_decay,
    _HP_model_chen,
)


def solve_evolution(
    NWN: NanowireNetwork, 
    t_eval: npt.NDArray,
    source_node: NWNNode | list[NWNNode], 
    drain_node: NWNNode | list[NWNNode], 
    voltage_func: Callable,
    window_func: Callable = None,
    tol: float = 1e-12,
    model: str = "default",
    solver: str = "spsolve",
    **kwargs
) -> tuple[OdeResult, list[NWNEdge]]:
    """
    Solve for the state variables `w` of the junctions of the given nanowire
    network at various points in time with an applied voltage.
    
    Parameters
    ----------
    NWN: Graph
        Nanowire network.

    t_eval : ndarray
        Time points to evaluate the nanowire network at. These should have
        units of `t0`.

    source_node : tuple, or list of tuples
        Voltage source nodes.

    drain_node : tuple, or list of tuples
        Grounded output nodes.

    voltage_func : Callable
        The applied voltage with the calling signature `func(t)`. The voltage 
        should have units of `v0`.

    window_func : Callable, optional
        The window function used in the derivative of `w`. The calling
        signature is `f(w)` where w is the array of state variables.
        The default window function is `f(w) = 1`.

    tol : float, optional
        Tolerance of `scipy.integrate.solve_ivp`. Defaults to 1e-12.

    model : {"default", "decay", "chen"}, optional
        Evolutionary model type. Default: "default".

    solver : str, optional
        Name of sparse matrix solving algorithm to use. Default: "spsolve".

    **kwargs
        Keyword arguments passed to the solver.
    
    Returns
    -------
    sol : OdeResult
        Output from `scipy.integrate.solve_ivp`. See the SciPy documentation
        for information on this output's formatting.

    edge_list : list of tuples
        List of the edges corresponding with each `w`.

    """
    if model not in ("default", "decay", "chen"):
        raise ValueError(f'Unsupported model type: model="{model}"')

    # Default window function
    if window_func is None:
        window_func = lambda x: 1

    # Get list of junction edges and state variable w
    edge_list, w0 = map(list, zip(*[
        ((u, v), w) for u, v, w in NWN.edges.data("w") if w is not None]
    ))
    
    # Get edge indices
    start_nodes, end_nodes = get_edge_indices(NWN, edge_list)
    
    # Get list of tau
    tau0 = [tau for _, _, tau in NWN.edges.data("tau") if tau is not None]

    # Get list of epsilon
    epsilon0 = [ep for _, _, ep in NWN.edges.data("epsilon") if ep is not None]

    # Define initial state and associated derivative
    if model == "default":
        _deriv = _HP_model_no_decay
        y0 = w0
    elif model == "decay":
        _deriv = _HP_model_decay
        y0 = w0
    elif model == "chen":
        _deriv = _HP_model_chen
        y0 = np.hstack((w0, tau0, epsilon0))
        if "sigma" not in NWN.graph.keys():
            raise AttributeError(
                "sigma, theta, and a must be set before using the Chen model.")

    # Solve the system of ODEs
    t_span = (t_eval[0], t_eval[-1])
    sol = solve_ivp(
        _deriv, t_span, y0, "DOP853", t_eval, 
        atol = tol, 
        rtol = tol,
        args = (NWN, source_node, drain_node, voltage_func, edge_list, 
            start_nodes, end_nodes, window_func, solver, kwargs)
    )

    # Update NWN to final state variables.
    if model == "default":
        set_state_variables(NWN, sol.y[:, -1], edge_list)
    elif model == "decay":
        set_state_variables(NWN, sol.y[:, -1], edge_list)
    elif model == "chen":
        w_list, tau_list, eps_list = np.split(sol.y, 3)
        set_state_variables(NWN, w_list[:, -1], 
            tau_list[:, -1], eps_list[:, -1], edge_list)

    return sol, edge_list


def set_state_variables(NWN: NanowireNetwork, *args):
    """
    Sets the given nanowire network's state variable. Can be called in the
    following ways:

        set_state_variables(NWN, w)
            where `w` is a scalar value which is set for all edges.

        set_state_variables(NWN, w, edge_list)
            where `w` is an ndarray and edge_list is a list. The `w` array
            contains the state variable for the corresponding edge in 
            `edge_list`.

        set_state_variables(NWN, w, tau)
            where `w` and `tau` are scalar values which is set for all edges.

        set_state_variables(NWN, w, tau, edge_list)
            where `w` and `tau` are ndarrays and edge_list is a list. The `w` 
            and `tau` arrays contains the state variables for the 
            corresponding edge in `edge_list`.

        set_state_variables(NWN, w, tau, epsilon)
            where `w`, `tau`, and `epsilon` are scalar values which is set for 
            all edges.

        set_state_variables(NWN, w, tau, epsilon, edge_list)
            where `w`,`tau`, and `epsilon` are ndarrays and edge_list is a list. 
            The `w`,`tau`, and `epsilon` arrays contains the state variables 
            for the corresponding edge in `edge_list`.

    Parameters
    ----------
    NWN: Graph
        Nanowire network. 

    *args
        See above.
    
    """
    # Only scalar w passed
    if len(args) == 1:
        if isinstance(args[0], Number):
            w = args[0]
            R = resist_func(NWN, w)

            attrs = {
                edge: {
                    "w": w, "conductance": 1 / R
                } for edge in NWN.edges if NWN.edges[edge]["type"] == "junction"
            }
            nx.set_edge_attributes(NWN, attrs)

        else:
            raise ValueError("Invalid arguments.")

    elif len(args) == 2:
        # vector w and edge_list passed
        if isinstance(args[0], np.ndarray) and isinstance(args[1], Iterable):
            w, edge_list = args
            R = resist_func(NWN, w)

            attrs = {
                edge: {
                    "w": w[i], "conductance": 1 / R[i]
                } for i, edge in enumerate(edge_list)
            }
            nx.set_edge_attributes(NWN, attrs)

        # scalar w and scalar tau passed
        elif isinstance(args[0], Number) and isinstance(args[1], Number):
            w, tau = args
            R = resist_func(NWN, w)

            attrs = {
                edge: {
                    "w": w, "conductance": 1 / R, "tau": tau
                } for edge in NWN.edges if NWN.edges[edge]["type"] == "junction"
            }
            nx.set_edge_attributes(NWN, attrs)
            NWN.graph["tau"] = tau

        else:
            raise ValueError("Invalid arguments.")

    elif len(args) == 3:
        # vector w, vector tau, and edge_list passed
        if (isinstance(args[0], np.ndarray) and 
            isinstance(args[1], np.ndarray) and 
            isinstance(args[2], Iterable)):
            w, tau, edge_list = args
            R = resist_func(NWN, w)

            attrs = {
                edge: {
                    "w": w[i], "conductance": 1 / R[i], "tau": tau[i]
                } for i, edge in enumerate(edge_list)
            }
            nx.set_edge_attributes(NWN, attrs)

        # scalar w, scalar tau, scalar epsilon passed
        if (isinstance(args[0], Number) and 
            isinstance(args[1], Number) and 
            isinstance(args[2], Number)):
            w, tau, epsilon = args
            R = resist_func(NWN, w)

            attrs = {
                edge: {
                    "w": w, "conductance": 1 / R,
                    "tau": tau, "epsilon": epsilon
                } for edge in NWN.edges if NWN.edges[edge]["type"] == "junction"
            }
            nx.set_edge_attributes(NWN, attrs)
            NWN.graph["tau"] = tau
            NWN.graph["epsilon"] = epsilon

        else:
            raise ValueError("Invalid arguments.")

    elif len(args) == 4:
        # vector w, vector tau, vector epsilon, and edge_list passed
        if (isinstance(args[0], np.ndarray) and 
            isinstance(args[1], np.ndarray) and
            isinstance(args[2], np.ndarray) and
            isinstance(args[3], Iterable)):
            w, tau, epsilon, edge_list = args
            R = resist_func(NWN, w)

            attrs = {
                edge: {
                    "w": w[i], "conductance": 1 / R[i], 
                    "tau": tau[i], "epsilon": epsilon[i]
                } for i, edge in enumerate(edge_list)
            }
            nx.set_edge_attributes(NWN, attrs)

        else:
            raise ValueError("Invalid arguments.")

    else:
        raise ValueError("Invalid number of arguments.")


def get_evolution_current(
    NWN: NanowireNetwork, 
    sol: OdeResult, 
    edge_list: list[NWNEdge], 
    source_node: NWNNode | list[NWNNode], 
    drain_node: NWNNode | list[NWNNode], 
    voltage_func: Callable,
    scaled: bool = False,
    solver: str = "spsolve",
    **kwargs
) -> npt.NDArray:
    """
    To be used in conjunction with `solve_evolution`. Takes the output from
    `solve_evolution` and finds the current passing through each drain node
    at each time step.

    The appropriate parameters passed should be the same as `solve_evolution`.

    Parameters
    ----------
    NWN : Graph
        Nanowire network.

    sol : OdeResult
        Output from `solve_evolution`.

    edge_list : list of tuples
        Output from `solve_evolution`.

    source_node : tuple, or list of tuples
        Voltage source nodes.

    drain_node : tuple, or list of tuples
        Grounded output nodes.

    voltage_func : Callable
        The applied voltage with the calling signature `func(t)`. The voltage 
        should have units of `v0`.

    scaled : bool, optional
        Scale the output by the characteristic values. Default: False.

    solver : str, optional
        Name of sparse matrix solving algorithm to use. Default: "spsolve".

    **kwargs
        Keyword arguments passed to the solver.

    Returns
    -------
    current_array: ndarray
        Array containing the current flow through each drain node. Each column
        corresponds to a drain node in the order passed.

    """
    # Get lists of source and drain nodes
    if isinstance(source_node, tuple):
        source_node = [source_node]
    if isinstance(drain_node, tuple):
        drain_node = [drain_node]
        
    # Preallocate output
    current_array = np.zeros((len(sol.t), len(drain_node)))

    # Loop through each time step
    for i in range(len(sol.t)):
        # Set state variables and get drain currents
        input_V = voltage_func(sol.t[i])
        set_state_variables(NWN, sol.y.T[i], edge_list)
        current_array[i] = solve_drain_current(
            NWN, source_node, drain_node, input_V, scaled, solver, **kwargs)
    
    return current_array.squeeze()


def get_evolution_node_voltages(
    NWN: nx.Graph, 
    sol: OdeResult, 
    edge_list: list[NWNEdge], 
    source_node: NWNNode | list[NWNNode],
    drain_node: NWNNode | list[NWNNode],
    voltage_func: Callable,
    solver: str = "spsolve",
    **kwargs
) -> npt.NDArray:
    """
    To be used in conjunction with `solve_evolution`. Takes the output from
    `solve_evolution` and finds the voltage of all nodes in the network at each
    time step.

    The appropriate parameters passed should be the same as `solve_evolution`.

    Parameters
    ----------
    NWN : Graph
        Nanowire network.

    sol : OdeResult
        Output from `solve_evolution`.

    edge_list : list of tuples
        Output from `solve_evolution`.

    source_node : tuple, or list of tuples
        Voltage source nodes.

    drain_node : tuple, or list of tuples
        Grounded output nodes.

    voltage_func : Callable
        The applied voltage with the calling signature `func(t)`. The voltage 
        should have units of `v0`.

    solver : str, optional
        Name of sparse matrix solving algorithm to use. Default: "spsolve".

    **kwargs
        Keyword arguments passed to the solver.

    Returns
    -------
    node_voltage_array: ndarray
        An `n` x `m` array containing the voltage of each node in the network,
        corresponding to the `n` time steps and `m` nodes in the network.

    """
    # Get lists of source and drain nodes
    if isinstance(source_node, tuple):
        source_node = [source_node]
    if isinstance(drain_node, tuple):
        drain_node = [drain_node]
        
    # Preallocate output
    node_voltage_array = np.zeros((len(sol.t), len(NWN.nodes)))

    # Loop through each time step
    for i in range(len(sol.t)):
        # Set state variables and get drain currents
        input_V = voltage_func(sol.t[i])
        set_state_variables(NWN, sol.y.T[i], edge_list)
        *node_voltage_array[i], I = solve_network(
            NWN, source_node, drain_node, input_V, "voltage", solver, **kwargs
        )
    
    return node_voltage_array.squeeze()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Functions to modify nanowire networks.
# 
# Author: Marcus Kasdorf
# Date:   July 26, 2021

from __future__ import annotations

import numpy as np
from shapely.geometry import LineString

from .typing import *
from typing import Iterable, TYPE_CHECKING
if TYPE_CHECKING:
    from .nanowire_network import NanowireNetwork

from .line_functions import find_line_intersects, add_points_to_line


def convert_NWN_to_MNR(NWN: NanowireNetwork):
    """
    Converts a JDA NWN to an MNR NWN in-place.

    Parameters
    ----------
    NWN : nx.Graph
        JDA nanowire network.

    """
    if NWN.graph["type"] == "MNR":
        print("Nanowire network already MNR.")
        return

    NWN.graph["type"] = "MNR"
    l0 = NWN.graph["units"]["l0"]
    rho0 = NWN.graph["units"]["rho0"]
    Ron = NWN.graph["units"]["Ron"]
    D0 = NWN.graph["units"]["D0"]
    A0 = D0*D0

    rho = NWN.graph["wire_resistivity"]
    D = NWN.graph["wire_diameter"]
    A = np.pi/4 * D*D

    for i in range(NWN.graph["wire_num"]):
        # Get the junctions for a wire
        junctions = NWN.edges((i,))

        # Get location of the junction for a wire
        junction_locs = {
            edge: NWN.graph["loc"][tuple(sorted([edge[0][0], edge[1][0]]))] for edge in junctions
        }

        # Add junctions as part of the LineString that makes up the wire
        NWN.graph["lines"][i], ordering = add_points_to_line(
            NWN.graph["lines"][i], junction_locs.values(), return_ordering=True
        )

        # If wire is electrode, move on to the next wire
        if (i,) in NWN.graph["electrode_list"]:
            continue

        # Split nodes into subnodes representing the junctions on the wires
        for j, (edge, loc) in enumerate(junction_locs.items()):
            # Find connecting node
            other_node = edge[~edge.index((i,))]

            # Get old edge attributes
            old_attributes = NWN.edges[edge]

            # Create the replacing MNR node and edge
            NWN.add_node((i, j), loc=loc, electrode=False)
            NWN.add_edge((i, j), other_node, **old_attributes)

        # Remove old JDA node, only if it had connections
        if len(junction_locs) >= 1:
            NWN.remove_node((i,))

        # Add edges between subnodes
        for ind, next_ind in zip(ordering, ordering[1:]):
            # Find inner-wire resistance
            L = NWN.nodes[(i, ind)]["loc"].distance(NWN.nodes[(i, next_ind)]["loc"])
            wire_conductance = (Ron * A0 * A) / (rho0 * l0 * rho * L * 1e3)

            # Add inner-wire edge
            NWN.add_edge(
                (i, ind), (i, next_ind), 
                conductance = wire_conductance, 
                capacitance = 0,
                type = "inner"
            )

    # Update index lookup
    NWN.graph["node_indices"] = {
        node: ind for ind, node in enumerate(sorted(NWN.nodes()))
    }


def add_wires(
    NWN: NanowireNetwork, 
    lines: list[LineString], 
    electrodes: list[bool], 
    resistance: float = None
) -> list[JDANode]:
    """
    Adds wires to a given nanowire network in-place. Returns the nodes of the 
    added wires in order.

    Currently, adding a wire that already exists breaks things.

    Also, adding wires to a MNR NWN does not work yet.

    Parameters
    ----------
    NWN : Graph
        Nanowire network to add wires to.

    lines : list of LineStrings
        A list of Shapely LineStrings, representing nanowires, to be added.

    electrodes : list of bool
        A list of boolean values specifying whether or not the corresponding
        nanowire in `lines` is an electrode.

    resistance : float, optional
        Junction resistances of the added wires.

    Returns
    -------
    new_wire_nodes : list of tuples
        List of the newly added nodes in the same order in `lines`.
    
    """
    if NWN.graph["type"] != "JDA":
        raise NotImplementedError("Only JDA is currently supported")

    new_wire_num = len(lines)

    if new_wire_num != len(electrodes):
        raise ValueError("Length of new lines list must equal length of electrode boolean list.")

    # Update wire number in NWN
    start_ind = NWN.graph["wire_num"]
    NWN.graph["wire_num"] += new_wire_num

    # Keep track of new wires
    new_wire_nodes = []

    # Add wires to NWN
    for i in range(new_wire_num):
        # Create new node
        NWN.graph["lines"].append(lines[i])
        new_wire_nodes.append((start_ind + i,))
        NWN.add_node(
            (start_ind + i,), 
            electrode = electrodes[i]
        )
        
        # Keep track of the electrodes
        if electrodes[i]:
            NWN.graph["electrode_list"].append((start_ind + i,))

        # Find intersects
        intersect_dict = find_line_intersects(start_ind + i, NWN.graph["lines"])
        
        # Add edges to NWN
        conductance = 1 / resistance if resistance is not None else NWN.graph["junction_conductance"]
        NWN.add_edges_from(
            [((key[0],), (key[1],)) for key in intersect_dict.keys()], 
            conductance = conductance,
            capacitance = NWN.graph["junction_capacitance"],
            type = "junction"
        )
        NWN.graph["loc"].update(intersect_dict)

        # Update index lookup
        NWN.graph["node_indices"].update({(start_ind + i,): start_ind + i})

    # Update wire density
    NWN.graph["wire_density"] = (NWN.graph["wire_num"] - len(NWN.graph["electrode_list"])) / NWN.graph["size"]

    # Clear current junction list
    del NWN.wire_junctions

    return new_wire_nodes


def add_electrodes(NWN: NanowireNetwork, *args)  -> list[JDANode]:
    """
    Convenience function for adding electrodes on the edges of a network 
    in-place. Returns the nodes of the added electrodes in order.

    Can be called in two ways:

        add_electrodes(NWN, *str)
            where *str are strings with values {"left", "right", "top, "bottom"}

        add_electrodes(NWN, *iterable)
            where *iterable are iterables where first entry is a string (as
            before) the second entry is number of electrodes on that side,
            and the third entry is the spacing between the electrodes,
            assumed to be in units of l0. An optional fourth entry can be
            given which is a list of offsets: a float value for each electrode.

    Parameters
    ----------
    NWN : Graph
        Nanowire network.

    *args
        See function description.

    Returns
    -------
    new_wire_nodes : list of tuples
        List of the newly added nodes. If strings were passed, the list
        follows the order passed. If iterables were passed, the list is
        ordered from left-to-right or bottom-to-top concatenated. 

    """
    length = NWN.graph["length"]
    width = NWN.graph["width"]
    line_list = []
    seen = []

    # Method one
    if all(isinstance(arg, str) for arg in args):
        for side in args:
            if side in seen:
                raise ValueError(f"Duplicate side: {side}")
            elif side == "left":
                line_list.append(LineString([(0, 0), (0, width)]))
            elif side == "right":
                line_list.append(LineString([(length, 0), (length, width)]))
            elif side == "top":
                line_list.append(LineString([(0, width), (length, width)]))
            elif side == "bottom":
                line_list.append(LineString([(0, 0), (length, 0)]))
            else:
                raise ValueError(f"Invalid side: {side}")
            seen.append(side)

    # Method two
    elif all(isinstance(arg, Iterable) for arg in args):
        for itr in args:
            # Unpack list
            if len(itr) == 3:
                side, num, spacing = itr
                offsets = [0.0] * num
            elif len(itr) == 4:
                side, num, spacing, offsets = itr
            else:
                raise ValueError("Invalid arguments.")

            # Cannot have the same side twice
            if side in seen:
                raise ValueError(f"Duplicate side: {side}")

            # Add electrodes
            elif side == "left":
                for i in range(num):
                    delta = offsets[i]
                    start = (i / num * width) + (spacing / 2)
                    end = ((i + 1) / num * width) - (spacing / 2)
                    line_list.append(LineString(
                        [(0, start + delta), (0, end + delta)]))
            elif side == "right":
                for i in range(num):
                    delta = offsets[i]
                    start = (i / num * width) + (spacing / 2)
                    end = ((i + 1) / num * width) - (spacing / 2)
                    line_list.append(LineString(
                        [(length, start + delta), (length, end + delta)]))
            elif side == "top":
                for i in range(num):
                    delta = offsets[i]
                    start = (i / num * length) + (spacing / 2)
                    end = ((i + 1) / num * length) - (spacing / 2)
                    line_list.append(LineString(
                        [(start + delta, width), (end + delta, width)]))
            elif side == "bottom":
                for i in range(num):
                    delta = offsets[i]
                    start = (i / num * length) + (spacing / 2)
                    end = ((i + 1) / num * length) - (spacing / 2)
                    line_list.append(LineString(
                        [(start + delta, 0), (end + delta, 0)]))
      
    else:
        raise ValueError("Arguments after NWN must be all strings or all lists.")
        
    # Add wires to the network
    new_wire_nodes = add_wires(NWN, line_list, [True] * len(line_list))
    return new_wire_nodes


def get_edge_indices(NWN: NanowireNetwork, edges: list[NWNEdge]) -> tuple[list[NWNNodeIndex], list[NWNNodeIndex]]:
    """
    Given a NWN and a list of edges, returns two lists: one of the indices of
    the first nodes in the input edge list, and one of the second.

    Parameters
    ----------
    NWN : Graph
        Nanowire Network.

    edges : list of tuples
        List of edges to find the indices of.

    """
    # JDA edge indices
    if NWN.graph["type"] == "JDA":
        start_nodes, end_nodes = map(list, 
            zip(*[(*n1, *n2) for n1, n2 in edges]))

    # MNR edge indices
    elif NWN.graph["type"] == "MNR":
        tmp = []
        for key in NWN.graph["node_indices"].keys():
            if len(key) == 2:
                tmp.append(key[1])
            else:
                tmp.append(0)

        node_start_index = np.where(np.asarray(tmp) == 0)[0]

        start_nodes, end_nodes = [], []
        for n1, n2 in edges:
            if len(n1) == 2:
                start_nodes.append(node_start_index[n1[0]] + n1[1])
            else:
                start_nodes.append(node_start_index[n1[0]])
            if len(n2) == 2:
                end_nodes.append(node_start_index[n2[0]] + n2[1])
            else:
                end_nodes.append(node_start_index[n2[0]])

    # Invalid NWN Type
    else:
        raise ValueError("Invalid NWN type.")

    return start_nodes, end_nodes


import pytest
from pathlib import Path

from randomnwn import *
from randomnwn.fromtext import *
from randomnwn.nanowire_network import NanowireNetwork


@pytest.fixture
def NWN_benchmark_JDA():
    # Get benchmark file
    current_path = Path(__file__).parent.resolve()
    benchmark = current_path.joinpath("test_networks/benchmark.txt")

    # Create nanowire network
    units = {"Ron": 20.0, "rho0": 22.63676, "D0": 60.0, "l0": 1.0}
    NWN = create_NWN_from_txt(str(benchmark), units=units)
    return NWN


@pytest.fixture
def NWN_benchmark_MNR(NWN_benchmark_JDA):
    NWN = NWN_benchmark_JDA
    convert_NWN_to_MNR(NWN)
    return NWN


@pytest.fixture
def NWN_test1():
    NWN = create_NWN(size=(8, 5), seed=123)
    add_electrodes(
        NWN, ["left", 2, 1, [-0.5, 0.5]], ["right", 2, 1, [-0.5, 0.5]]
    )
    return NWN


@pytest.mark.parametrize("NWN", ["NWN_benchmark_JDA", "NWN_benchmark_MNR", "NWN_test1"])
def test_NWN_edge_indices(NWN, request):
    NWN = request.getfixturevalue(NWN)

    # Get edge indices
    start1, end1 = map(np.asarray, get_edge_indices(NWN, NWN.wire_junctions))
    start2, end2 = np.asarray(NWN.get_index_from_edge(NWN.wire_junctions)).T

    # Check if the edge indices are the same
    assert np.all(start1 == start2)
    assert np.all(end1 == end2)


@pytest.mark.parametrize("NWN_input", ["NWN_benchmark_JDA", "NWN_benchmark_MNR"])
@pytest.mark.parametrize("voltage_func", [
    lambda t: 1 + 0*t,
    lambda t: 2 + 1*t,
])
@pytest.mark.parametrize("window_func", [
    lambda w: 1,
    lambda w: w * (1 - w),
])
def test_new_api_evolution(NWN_input, voltage_func, window_func, request):
    # Common vars
    max_time = 10
    n_time = 100
    t_eval = np.linspace(0, max_time, n_time)

    # Evolution with new API
    NWN: NanowireNetwork = request.getfixturevalue(NWN_input)
    source, drain = NWN.electrodes

    NWN.state_vars = ["w"]
    NWN.set_state_var("w", 0.05)

    sol1 = NWN.evolve(
        "default", t_eval, source, drain, voltage_func,
        window_func, ivp_options={"rtol": 1e-7, "atol": 1e-7}
    )

    # Evolution with old API
    NWN: NanowireNetwork = request.getfixturevalue(NWN_input)
    source, drain = NWN.electrodes

    set_state_variables(NWN, 0.05)

    sol2, _ = solve_evolution(
        NWN, t_eval, source, drain, voltage_func, window_func, 1e-7, "default"
    )

    # Compare the results
    assert np.all(sol1.y == sol2.y)



"""
Random NWNs package.

Can be imported using either:

    >>> import randomnwn as rnwn

or:

    >>> from randomnwn import *

"""
from .version import __version__

from .nanowire_network import create_NWN

from .nanowires import (
    convert_NWN_to_MNR,
    add_wires,
    add_electrodes,
    get_edge_indices,
)

from .line_functions import (
    create_line,
    find_intersects,
    find_line_intersects,
    add_points_to_line,
)

from .calculations import (
    get_connected_nodes,
    create_matrix,
    solve_network,
    solve_drain_current,
    solve_nodal_current,
    solve_edge_current,
    scale_sol,
)

from .plotting import (
    plot_NWN,
    draw_NWN,
)

from .dynamics import (
    resist_func,
    solve_evolution,
    set_state_variables,
    get_evolution_current,
    get_evolution_node_voltages,
)

from .units import (
    get_units
)

from ._models import (
    set_chen_params,
)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Functions to plot nanowire networks.
# 
# Author: Marcus Kasdorf
# Date:   July 8, 2021

import numpy as np
import networkx as nx
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import matplotlib as mpl

from matplotlib.figure import Figure
from matplotlib.axes import Axes


def plot_NWN(
    NWN: nx.Graph, 
    intersections: bool = True, 
    rnd_color: bool = False,
    scaled: bool = False,
    grid: bool = True,
    xlabel: str = "",
    ylabel: str = "",
) -> tuple[Figure, Axes]:
    """
    Plots a given nanowire network and returns the figure and axes.

    Parameters
    ----------
    NWN : Graph
        Nanowire network to plot.

    intersections : bool, optional
        Whether or not to scatter plot the intersections as well.
        Defaults to true.

    rnd_color : bool, optional
        Whether or not to randomize the colors of the plotted lines.
        Defaults to false.

    scaled: bool, optional
        Whether or not to scale the plot by the characteristic values of the
        given nanowire network. Defaults to False.

    grid: bool, optional
        Grid lines on plot. Defaults to true.

    xlabel: str, optional
        x label string.

    ylabel: str, optional
        y label string.

    Returns
    -------
    fig : Figure
        Figure object of the plot.

    ax : Axes
        Axes object of the plot.

    """
    fig, ax = plt.subplots(figsize=(8, 6))
    l0 = NWN.graph["units"]["l0"]

    # Plot intersection plots if required
    if intersections:
        ax.scatter(
            *np.array([(point.x, point.y) for point in NWN.graph["loc"].values()]).T, 
            zorder=10, s=5, c="blue"
        )

    # Defaults to blue and pink lines, else random colors are used.
    if rnd_color:
        for i in range(NWN.graph["wire_num"]):
            ax.plot(*np.asarray(NWN.graph["lines"][i].coords).T)
    else:
        for i in range(NWN.graph["wire_num"]):
            if (i,) in NWN.graph["electrode_list"]:
                ax.plot(*np.asarray(NWN.graph["lines"][i].coords).T, c="xkcd:light blue")
            else:
                ax.plot(*np.asarray(NWN.graph["lines"][i].coords).T, c="pink")

    # Scale axes according to the characteristic values
    if scaled:
        ax.xaxis.set_major_formatter(
            ticker.FuncFormatter(lambda x, pos: f"{x * l0:.3g}")
        )
        ax.yaxis.set_major_formatter(
            ticker.FuncFormatter(lambda y, pos: f"{y * l0:.3g}")
        )

    # Other attributes
    if grid: 
        ax.grid(alpha=0.25)
    if xlabel: 
        ax.set_xlabel(xlabel)
    if ylabel: 
        ax.set_ylabel(ylabel)

    plt.show()
    return fig, ax


def draw_NWN(
    NWN: nx.Graph, 
    figsize: tuple = None,
    font_size: int = 8,
    node_labels: np.ndarray = None,
    fmt: str = ".2f",
    edge_colors: np.ndarray = None,
    cbar_label: str = "Colorbar",
    cmap = plt.cm.RdYlBu_r,
) -> tuple[Figure, Axes]:
    """
    Draw the given nanowire network as a networkx graph. JDA drawing is more
    detailed as nodes can be given spacial locations. With MNR drawing, nodes
    will have random locations.

    Parameters
    ----------
    NNW : Graph
        Nanowire network to draw.

    figsize : tuple, optional
        Figure size to be passed to `plt.subplots`.

    font_size : int, optional
        Font size to be passed to `nx.draw`.

    node_labels : ndarray, optional
        If supplied, these values will be display as node labels
        instead of the names of the nodes.

    fmt : str, optional
        String formatting for node labels. Only used if sol is passed.
        Default: ".2f".

    edge_colors : ndarray, optional
        List of values to color the edges. Edges are assumed to be in the
        same order as `NWN.edges`.

    cbar_label : str, optional
        Label for the colorbar.

    cmap : colormap, optional
        Matplotlib color map to use for the edges.

    Returns
    -------
    fig : Figure
        Figure object of the plot.

    ax : Axes
        Axes object of the plot.

    """
    fig, ax = plt.subplots(figsize=figsize)

    if NWN.graph["type"] == "JDA":
        kwargs = dict()

        # Nodes are placed at the center of the wire
        kwargs.update({
            "pos": {(i,): np.asarray(*NWN.graph["lines"][i].centroid.coords) 
                for i in range(NWN.graph["wire_num"])}
        })

        # Label node voltages if sol is given, else just label as nodes numbers
        if node_labels is not None:
            kwargs.update({
                "labels": {(key,): f"{value:{fmt}}" 
                    for key, value in zip(range(NWN.graph["wire_num"]), node_labels)}
            })
        else:
            kwargs.update({
                "labels": {(i,): i for i in range(NWN.graph["wire_num"])}
            })

        # Add edges colors if weights are passed
        if edge_colors is not None:
            kwargs.update({
                "edgelist": NWN.edges, 
                "edge_color": edge_colors, 
                "edge_cmap": cmap
            })

            # Add a colorbar to the network plot
            norm = mpl.colors.Normalize(
                vmin=np.nanmin(edge_colors), vmax=np.nanmax(edge_colors))

            cax = fig.add_axes([0.95, 0.2, 0.02, 0.6])
            cb = mpl.colorbar.ColorbarBase(cax, norm=norm, cmap=cmap)
            cb.set_label(cbar_label)

        else:
            kwargs.update({"edge_color": "r"})

        # Add node formatting
        kwargs.update({"ax": ax, "font_size": font_size, "node_size": 40})

        nx.draw(NWN, **kwargs)

    elif NWN.graph["type"] == "MNR":
        kwargs = {}
        if node_labels is not None:
            labels = {node: f"{value:{fmt}}" for node, value in zip(sorted(NWN.nodes()), node_labels)}
            kwargs.update({"labels": labels})
        else:
            kwargs.update({"with_labels": True})

        nx.draw(NWN, ax=ax, node_size=40, font_size=font_size, edge_color="r", **kwargs)

    else:
        raise ValueError("Nanowire network has invalid type.")

    plt.show()
    return fig, ax




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Typing definitions for randomnwn.
# 
# Author: Marcus Kasdorf
# Date:   July 8, 2024

from typing import TypeAlias

JDANode: TypeAlias = tuple[int]
MNRNode: TypeAlias = tuple[int, int]
NWNNode: TypeAlias = JDANode | MNRNode
NWNEdge: TypeAlias = tuple[NWNNode, NWNNode]
NWNNodeIndex: TypeAlias = int
NWNEdgeIndex: TypeAlias = tuple[NWNNodeIndex, NWNNodeIndex]


from setuptools import setup

namespace = {}
with open("randomnwn/version.py", "r") as f:
    exec(f.read(), namespace)

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="randomnwn",
    version=namespace["__version__"],    
    description="Modelling and analyzing random nanowire networks in Python.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/marcus-k/Random-NWNs",
    author="Marcus Kasdorf",
    author_email="marcus.kasdorf@ucalgary.ca",
    license="GNU General Public License v3",
    packages=["randomnwn"],
    python_requires=">=3.10",
    install_requires=[
        "matplotlib>=3.8.4",
        "networkx>=3.1",
        "numpy>=1.26.4",
        "scipy>=1.13.0",
        "Shapely>=2.0.4,<3",
    ],
)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Functions for random lines.
# 
# Author: Marcus Kasdorf
# Date:   May 20, 2021

import numpy as np
from numpy.random import uniform
from shapely.geometry import LineString, Point


def create_line(length=1, xmin=0, xmax=1, ymin=0, ymax=1, rng=None) -> LineString:
    """
    Generate random lines with random orientations with midpoints ranging from 
    area from ``xmin`` to ``xmax`` and from ``ymin`` to ``ymax``.

    Parameters
    ----------
    length : float
        Length of line
    
    xmin : float
        Minimum x coordinate midpoint.

    xmax : float
        Minimum x coordinate midpoint.

    ymin : float
        Minimum y coordinate midpoint.

    ymax : float
        Minimum y coordinate midpoint.

    rng : Generator
        Generator object usually created from ``default_rng`` from 
        ``numpy.random``. A seeded generator can be passed for consistent 
        random numbers. If None, uses the default NumPy random functions.

    Returns
    -------
    out : LineString
        LineString of the generated line.

    """
    if rng is not None:
        xmid, ymid, angle = rng.uniform(xmin, xmax), rng.uniform(ymin, ymax), rng.uniform(0, np.pi)
    else:
        xmid, ymid, angle = uniform(xmin, xmax), uniform(ymin, ymax), uniform(0, np.pi)

    xhalf, yhalf = length / 2 * np.cos(angle), length / 2 * np.sin(angle)

    xstart, xend = xmid - xhalf, xmid + xhalf
    ystart, yend = ymid - yhalf, ymid + yhalf

    out = LineString([(xstart, ystart), (xend, yend)])
    return out


def find_intersects(lines: list) -> dict[tuple[int, int], Point]:
    """
    Given a list of LineStrings, finds all the lines that intersect and where.

    Parameters
    ----------
    lines : list of LineStrings
        List of the LineStrings to find the intersections of.

    loc : bool
        Whether or not to return the intersect locations. Defaults to false.

    Returns
    -------
    out : dict
        Dictionary where the key is a tuple of the pair of intersecting lines 
        and the value is the intersection locations.

    """
    out = {}

    for i, j in zip(*np.triu_indices(n=len(lines), k=1)):
        # Check for intersection first before calculating it
        if lines[i].intersects(lines[j]):
            out.update({(i, j): lines[i].intersection(lines[j])})

    return out


def find_line_intersects(ind: int, lines: list[LineString]) -> dict[tuple[int, int], Point]:
    """
    Given a list of LineStrings, find all the lines that intersect
    with a specified line in the list given by the index ``ind``.

    """
    out = {}

    for j in range(len(lines)):
        # Skip intersection with the line itself
        if ind == j:
            continue
        
        # Checking if these's an intersection first is faster
        if lines[ind].intersects(lines[j]):
            if ind < j:
                out.update({(ind, j): lines[ind].intersection(lines[j])})
            else:
                out.update({(j, ind): lines[ind].intersection(lines[j])})

    return out


def add_points_to_line(line: LineString, points: list[Point], return_ordering=False):
    """
    Given a list of points and a line, add the projected points to the line.

    See general form at: 
    https://stackoverflow.com/questions/34754777/shapely-split-linestrings-at-intersections-with-other-linestrings

    """
     # First coords of line (start + end)
    coords = [line.coords[0], line.coords[-1]]

    # Add the coords from the points
    coords += [(p.x, p.y) for p in points]

    # Calculate the distance along the line for each point
    dists = [line.project(Point(p)) for p in coords]

    # Sort the coords based on the distances
    coords, ordering = map(list, zip(
        *[(point, ind) for _, point, ind in sorted(zip(dists, coords, range(len(coords))))]
    ))

    ordering.remove(0); ordering.remove(1)
    ordering = [ordering[i] - 2 for i in range(len(ordering))]

    # Overwrite old line
    line = LineString(coords)

    if return_ordering:
        return line, ordering
    else:
        return line
    




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Characteristic units for a nanowire network.
# 
# Author: Marcus Kasdorf
# Date:   July 1, 2021

def get_units(new_units: dict[str, float] = None) -> dict[str, float]:
    """
    Returns the characteristic units for a nanowire network.

    Parameters
    ----------
    new_units : dict, optional
        Dictionary of any custom units to use. Only base units can be altered.

    Returns
    -------
    units : dict
        Dictionary of characteristic units.
    
    """
    if new_units is None:
        new_units = dict()

    # Base units
    units = {               # Unit, Description
        "v0": 1.0,          # V, Voltage
        "Ron": 10.0,        # Ω, ON junction resistance
        "l0": 7.0,          # μm, Wire length
        "D0": 50.0,         # nm, Wire diameter
        "w0": 10.0,          # nm, Junction length (2x Wire coating thickness)
        "rho0": 22.6,       # nΩm, Wire resistivity
        "mu0": 1e-2,        # μm^2 s^-1 V^-1, Ion mobility
        "Roff_Ron": 160     # none, Off-On Resistance ratio
    }

    # Add any custom units
    units.update(new_units)

    # Derived units
    units["i0"] = units["v0"] / units["Ron"]                    # A, Current
    units["t0"] = units["w0"]**2 / (units["mu0"] * units["v0"])  # μs, Time

    return units

